(*! What follows is a program written in Vector Pascal.
 The text in roman font that follows
are comments. The program code is generally in san-serif font, and
the whole, is in the literate programming output format generated
by the compiler. *)
program excavate;
 
(*! 
The objective of the program is to solve Kantorovich's soil excavation
problem by his method of resolving multipliers. It starts out from
the data provided in Table \ref{excavator}. \include{table5}
In the table the {\em norms} for the excavator types are shown
in italics. In Soviet parlance, a norm appears to have meant the
expected output per hour of the a machine applied to a particular type of
work. Thus Excavator A can dig out 105 m$^3/$hr of soil type I,
56 m$^3/$hr of soil of type II etc.#


The objective is given in the last column : $20,000m^3$ of each type
of soil.

Following the columns of the norms Kantorovich gives the optimal allocation
of machine times to activities to minimise overall time taken to do the digging.
The program will reproduce this result by applying his method of resolving
multipliers or objectively determined valuations.
We first introduce our domain of discourse : the types of soil, the types of 
machine and the units of measurement we are using.
*)
type 
     soil = (I,II,III);
     excavator = (A,B,C);
     units =(hr, meter);
     
    
 (*! Now we introduce the dimensions in which volume, time and norms are specified.
 For instance norms are real numbers denoting
 cubic meter per hour. The word {\sf pow} in what follows means raised to the power, and * is
the multiplication operator. 
  
*)
 type 
	volume =  real of meter pow 3;
	duration = real of hr;
	norm = real of meter pow 3 * hr pow -1;
const
	hour  : duration =1.0;
	cubicmeter : volume =1.0;
	epsilon =0.001;
(*! The production norms for the machines working on each kind of soil and targets for soil to be moved 
are copied from Kantrovichs Table 5  and stored in an appropriate matrix
  called norms, and a vector called targets.  *)
const
	norms : array[soil,excavator] of norm =
						((105, 107, 64),
						( 56, 66,38),
						( 56, 83, 53));
        identity: array[soil,soil] of real =
						((1, 0, 0),
						( 0, 1, 0),
						( 0, 0, 1));
        target :   volume = 20000 ;
        
(*! We now introduce the variables of the problem: a matrix x which will encode  the time each machine
    spends on each type of soil; L, a vector of objectively determined valuations of different soils.
    The standardised output of each machine for each soil type is obtained by applying resolving multipliers
    to the soil types.
*)
var x: array[soil,excavator ] of duration;
    dx: duration;
    L : array [soil] of real; { resolving multipliers}
    standardisedoutput: array[ excavator,soil] of norm;    
    outputs : array[soil] of norm;  
    totals: array[ soil ] of volume;
    ok:boolean;  
    greatestsoil,leastsoil ,deltam: volume;
    best:norm;  
    e, Scoop:excavator; { }
    s,least,m,j: soil; 
    lambda,f:real; 
    count:integer;
    equated: array[soil] of real;
    procedure ComputeTotalsEtc;
	(*! Work out how much is being produced, which soil is  being produced least and which soils outputs are equals to this, *) 
     var d :array[soil]of real;
     begin
			totals := \+ (norms * x);
			leastsoil := \min totals;
			(*! Find the soil that is least produced. *)
			for s:= I to III do 
				if totals[s]=leastsoil then least:=s;
			(*! Find the ones on the plan ray *)
			d:=(totals-leastsoil)/cubicmeter;
			equated := if (d<epsilon)and(d> -epsilon) then 1.0 else 0.0;
		{		writeln('t',totals/cubicmeter);
				writeln('d',d);
				writeln('e',equated,'x',x/hour);
				writeln('least',least);
			write('press return');readln;}
     end;
    function marginalgain( s :soil; d :excavator):volume;
    (*! This computes the marginal gain, under the weighting imposed by the current resolving multipliers, of a
        small shift of the digger d's time to the specified soil type s. We compute the effect of 
        multiplying all current time allocations to $1-\epsilon$ whilst increasing the allocation of time
        to soil s by $\epsilon$. The assumption here is that for now each machine has only one hour to allocate. *)
     const epsilon =0.001;
     var currentoutput:volume;
     begin { marginal gain}
     
        {write('marginalgain(',s,',',d,')');}
		currentoutput :=  \+ x[d ]*norms[iota 0,d ] ;
		marginalgain :=  ((epsilon*hour ) * norms[s,d])-  epsilon *currentoutput ;
		writeln((((epsilon*hour ) * norms[s,d])-  epsilon *currentoutput)/cubicmeter );
     end;
     
     function mainSoilProducedBy(e:excavator):soil;
     (*! This determines which soil excavator e produces the most of. *)
     var v:volume;j,s:soil;
     begin
		v:=0 * cubicmeter;
        s:=i;
		for j:= I to III do
		begin
			if v< norms[j,e]*x[j,e] then
			begin
				v:=norms[j,e]*x[j,e];
				s:=j;
			 
			end;
		end;
	{	writeln('main output produced by',e,' is ',s);}
		mainSoilProducedBy:=s;
     end;
      
     function findScoop(var s:soil):excavator;
     (*!  search all soils who are not yet reaching target to find a pair of  machine and soil where
     shifting machine resources to producing that soil would produce the best result . Return the machine as the function
     result and the soil in the parameter s *)
     var
	gain:volume;j,m:soil;
	d,Scoop:excavator;  
     begin {find Scoop} 
      {  writeln('in find scoop equated =');write(equated);{}
		Scoop := A;
		gain := (-maxint)*cubicmeter;
		for d:= A to C do
		for j:= I to III do
		if equated[j]>0 then		{find a technique that is doing ok that potentially could have resource reallocated }
		begin
			m:=mainsoilproducedby(d); 			 
			 writeln(d,j,m);
			if marginalgain(j,d)>gain then
			begin
				if equated[m]<1 then
				begin 
					gain:= marginalgain(j,d);
				{	writeln('gain',gain/cubicmeter,',j',j,',d',d,',m',m);{}
					Scoop:= d; s:=j;
				end;
			end;
		end;
		findScoop:= Scoop;
     end;
    
begin {Kantorovich algorithm}
	L:=1; { initialise multipliers}
	ok:=false; count:=0;f:=0.3;
        (*! Iterate the following steps until we have a satisfactory answer. *)
	while not ok do 
	begin
		x:=0*hour ; {default assumption }
		(*!     Use the L    to get a standardised performance for each machine. *)
		standardisedoutput := trans(norms *   trans L);
		(*! For each machine find the soil for which it has the best performance *)
		for e := A to C do
		begin
		(*! find the best performance of the machine on any soil *)
			outputs:= standardisedoutput[e];
			best := \max outputs;
			(*! set each machine to work on the soil it is best at *)
			for  s := I to III do 
				if standardisedoutput[e,s]=best then x[s,e]:=  hour ;
		end;
		totals := \+ (norms * x);
		greatestsoil := \max totals;
		leastsoil := \min totals;
		if leastsoil<=0.0*cubicmeter then
		begin
		(*! check if any soil has a zero output and raise its value if it has*)
		for s:= I to III do 
			if totals[s]<greatestsoil then 
				L[s]:= L[s]*(1.02+(ord(s)/10))
		end
		else ok:=true;
	end;
	count:=0;
	(*! At this point our estimate of the resolving multipliers is accurate enough to ensure
            that some of each soil is now being moved, but we have not yet met the requirement
            that the same amount of each soil must be moved. We now try to get a more precise
            estimate of the resolving multipliers and in the process we adjust the amounts
            of each soil being moved. It is important to note at this point that any further
            adjustments must come by de-specialising some of the excavators so that they move
            more than one soil type. The resolving multipliers have until now been  used to weight the outputs
            of different soil types in order to assign each digger to the soil it is best suited to.
            If a machine is no longer specialised, that is if it moves more than one soil, then
            the weights must be such that it is no longer {\em best} at one particular soil type.
            The multipliers must be set so that the marginal weighted output of the excavator on
            any of the soils on which it is employed are the same. Thus if a machine $k$ is employed
            on two soils $i,j$ then {\sf standardisedoutput[k,j]=standarisedoutput[k,i]}. 

		In turn this implies that for any machine that is employed to move two soils
            the ratio of the resolving multipliers must be the inverse of the ratio of the norms.

	    The algorithm will work soil a time bringing ever more soil outputs into equality.
            We define the set of soils whose outputs has been brought into equality as the equated set.

	    For those soils in the equated set, the resolving multipliers of the soils will have been corrected
            so that for any machine moving more than one soil they stand in inverse ratio to
            that digger's norms.
	 *)
	 
	computeTotalsEtc; 
	repeat 			 
			(*! Find which machine  is 2nd best at producing soil least in the non equated set  under current resolving multipliers. Call this machine Scoop. *)
			 Scoop := findScoop(least);
		 
			 m := mainSoilProducedBy(Scoop); 
			 (*! Adjust the resolving multiplier ratio between Scoops soil and the least produced soil to
                            ratio of Scoops norms. *)
			L [m]:=L [least]*norms[least,scoop]/norms[m,scoop];
			(*! It is now necessary to reduce the ouput of scoop on scoops soil and increase it
                            on the least produced soil. It is necessary to compute how much to reduce scoops soil by.
                            The resolving multipliers give us substitution ratios between different soil outputs.
                            Suppose that we want to reduce output of soil $m$ by one unit and increase the output
                            of soil $j$, the increase in $j $ we get is
				 $$ \Delta_j=\frac{L_m}{L_j}$$.
                            If we want to reduce the output of $i$  and we have two other soils $j,k$ which
                            we want to increase  equally then we have 
				$$\Delta_k=\Delta_j$$ 
				where $\Delta_x$ means change in $x$ and
                            $$ - \Delta_m L_m = \Delta_k L_k + \Delta_j L_j=\Delta_k(L_j+L_k)$$.
				Thus $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m} $$
                            Let $C_m, C_j, C_k$ be the current outputs of each soil;  given that $C_j=C_k$ 
                            we have to chose the $\Delta$s so that $$ C_m+\Delta_m = C_j+\Delta_j=C_k+\Delta_k$$
                            It follows that $$C_m-\Delta_k\frac{L_j+L_k}{L_m} =C_k+\Delta_k$$
				and$$C_m-C_k=\Delta_k\frac{L_j+L_k}{L_m} +\Delta_k=\Delta_k(1+\frac{L_j+L_k}{L_m})$$
			   	so$$\Delta_j= \frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}$$
			We next compute the reduction to be made in soil $m$ from the formula $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m}$$
			substitiuting we get
			$$\Delta_m= -\frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}(\frac{L_j+L_k}{L_m})$$
			    Translating this to the variables used in the program we have:
                         *)
			j:=least;
		
			lambda :=  (L.equated)/L[m];
				
			deltam :=  (-1)*lambda*(totals[m]-totals[j])/(1+lambda) ; 

			(*!  Note that in the line above we are generalising the term $L_j+L_k$ to an aribtrary number
				 of multipliers ( 1 or 2  in this program ) by computing the inner product between the
				 equated vector and the multipliers. This works because the equated vector has a 1 for
				 all soils in the equated set.
    We now compute the change in duration that Scoop spends on its best soil (dx) by 
				scaling deltam by Scoops norm for soil m.
			 *)
	        dx := deltam/norms[m,Scoop];

   			x[m,Scoop]:=x[m,Scoop] +dx;
   			

			(*! reallocate this time to Scoops best soil in the equated set which we will now call $j$ *)
			best:=norms[I,Scoop]*0;j:=I;
			for s:=I to III do 
				if norms[s,Scoop] * equated[s] >best then
				begin   
					best:= norms[s,Scoop];
					j:=s;
				end;
		 
			x[j,Scoop]:= x[j,Scoop]-dx;
			
           	computeTotalsEtc;	  
			count:=count+1;
	until ((\+ equated) = 3) or (count>10);
	writeln('answer arrived at after ',count,' trys');
	writeln('intensities');
	writeln(x/hour);
	writeln (' allocation ' );
	
	writeln ( (x* target/totals[I]) /hour);
end {Kantorovich algorithm}.
