(*! What follows is a program written in Vector Pascal.
 The text in roman font that follows
are comments. The program code is generally in san-serif font, and
the whole, is in the literate programming output format generated
by the compiler. *)
program genKantor;
 
(*! 
* 
The programme has as its objective to generalise the algorithm of 
Kantorovich that he first presents in the context of excavators
working on soils of different types, so that it will work for
a general plan optimisation for the whole economy  
by his method of resolving multipliers. It starts out from
the data provided in an i/o Table.  
In the table the {\em techniques} for for different production techniques are shown
in italics. In this context a technique  means the
expected output per year of the technique. In the case of Kantorovich's
original algorithm the techniques were represented as a matrix of
outputs of different soils and machines. In an input output table
a technique is conventionally represented by a column (Leontief notation)
or a row (Sraffa notation) of a matrix. 

We can convert the Kantorovich table $$
\begin{array}{ccc}105& 107& 64\\
						 56& 66&38\\
						 56& 83& 53\end{array}$$
Where each row corresponds to a soil type and each column to
an excavator into an extended Sraffa format as 
$$
\begin{array}{cccccc}
105&0&0&-1&0&0\\
107&0&0& 0&-1&0\\
64&0&0& 0&0&-1\\
0&56&0& -1&0&0\\
0&66&0& 0&-1&0\\
0&38&0&  0&0&-1\\
0&0&56& -1&0&0\\
0&0&83& 0&-1&0\\
0&0&53&  0&0&-1
\end{array}
$$
where the first 3 columns correspond to the outputs of different soils
and the next 3 to the outputs of excavators. The outputs of excavators 
are negative since a given process uses up an excavator.

The objective was given, by Kantorovich  in a final  column  of his table,
in our format we would have a row vector of final outputs, and a second
row vector of resources available at the start of the period.

To copy Kantorovich target of 20000 meters of soil for each type
of soil we would have a target row vector
$$
\begin{array}{cccccc}
 20000&20000&20000&0&0&0
\end{array}
$$
And an initial resource vector of 
$$
\begin{array}{cccccc}
0&0&0&1&1&1
\end{array}
$$
indicating we have 1 machine of each type.

My first step in developing the programme in this new format was to 
take the programme that I had previously published in Intervention, and
to reformulate it in this more general form. The aim is to check that
given the new data format the algorithm still gives the same answers
as Kantorovich. Following on from that we can provide a more
general version that will work on arbitrary sized input output
tables.

Following the columns of the techniques Kantorovich gives the optimal allocation
of machine times to activities to minimise overall time taken to do the digging.
The program will reproduce this result by applying his method of resolving
multipliers or objectively determined valuations.
We first introduce our domain of discourse resources, techniques and units of  usage and output. We simply
number the resources 0 to 5 with resources in this case including soil
and machines.
*)
const 
      
      maxtechnique =18;
      maxoutput = 2;
      alpha=10;
      techniqueN=maxtechnique;
	  epsilon =0.001;
type 
     flow = (dest,src);
     resource = (I,II,III,A,B,C);
const
     resource0:resource=I;  { first entry in US IO table is resource0 }
     
     resourceM:resource=C; { last entry is resourceM }
type
     technique = alpha .. techniqueN;
     units =real;
     duration = real;
     resourcevec = array[resource] of units;
     (*! A technology is defined as a flow of  src inputs to a flow of dest outputs,
     with each flow being a vector. *)
     technology = array [flow] of resourcevec;
     volume=real;
 
(*! The production techniques for the machines working on each kind of soil and targets for soil to be moved 
are copied from Kantrovichs Table 5  and stored in an appropriate matrix
  called techniques, and a vector called targets.  *)
const techniques : array[technique ] of technology =
       (((105,0,0,0,0,0),(0,0,0, -1,0,0)),
		( (107,0,0,0,0,0),(0,0,0, 0,-1,0)),
		((64,0,0,0,0,0),(0,0,0, 0,0,-1)),
		((0,56,0,0,0,0),(0,0,0, -1,0,0)),
		((0,66,0,0,0,0),(0,0,0, 0,-1,0)),
		((0,38,0,0,0,0),(0,0,0,  0,0,-1)),
		((0,0,56,0,0,0),(0,0,0, -1,0,0)),
		((0,0,83, 0,0,0),(0,0,0,0,-1,0)),
		((0,0,53, 0,0,0),(0,0,0, 0,0,-1)));
const	produces:array[technique] of resource=(0,0,0,1,1,1,2,2,2);
		 
const     target : resourcevec=(20000,20000,20000,0,0,0) ;
(*! For our starting resources we allow sufficient hours of time on each machine
that we know that the desired target of Kantorovich can be reached. Later we
can try altering the resources to see how that affects the final outputs. *)
     startingresource : resourcevec =(0,0,0,1,1,1);
        
(*! We now introduce the variables of the problem: a vector x which will encode  the time  for which each technique
    is operated, this corresponds in Kantorovich's case to the time a given type of machine spends on a given soil. 
    Later it may have a more general impliation. Next  L, a vector of objectively determined valuations  of resources\footnote{
    Presumably Kantorovich labels his multipliers L because of their similarity to Lagrangian multipliers}.
    
    The standardised output of each machine for each soil type is obtained by applying resolving multipliers
    to the soil types.
*)
var x: array[technique ] of duration;
    targetintensity:real;
    achievableintensities:resourcevec;
    dx: duration;
    L : array [resource] of real; { resolving multipliers}
    standardisedoutput: array[ technique] of resourcevec;    
    available: resourcevec;
    outputs : array[technique] of real;  
    totals: array[ resource ] of real;
    ok:boolean;  
    greatestProducedOutput,leastProducedOutput ,deltam: real;
    best:real;  
    e, e2,Scoop:resource; { }
    s, j: resource; 
    weightedTotalOutput:real;{Output weighted by L}
    lambda,f:real; 
    least,m,jt,st:technique;
    count,k:integer;
    equated: array[resource] of real;
    function marginalgain( s :technique; d :resource):volume;
    (*! This computes the marginal gain,   of a
        small shift of the resource  d  time to the specified technique type s. We compute the effect of 
        multiplying all current time allocations to $1-\epsilon$ whilst increasing the allocation of time
        to resource s by $\epsilon$.   *)
    
     var currentoutput, currentuse, shift, marginalIncreaseinS:real;
         valueofincrease,valueofloss:real;
         t:technique;
     begin { marginal gain}
      
        (*! check if technique s uses resource d *)
        if techniques[s,src,d] =0 then
        begin
           
           marginalgain:= - maxint
        end
        else begin
	     (*! get total current use of resource d *)
	        currentuse := \+ x* techniques[iota 0, src, d];{ current use is negative !}
	        shift:= epsilon * currentuse;                  { shift will be negative too}
	        marginalIncreaseInS:= shift* techniques[s,dest,produces[s]]/techniques[s,src,d];{ marginal increase will be +ve}
	        
			 
			valueofincrease:= marginalIncreaseInS * L[produces[s]];
			(*! we next have to work out the total value of output currently produced in all sectors by d *)
			valueofloss:=0;
			for t:= alpha to techniqueN do
		 	 if techniques[t,src,d]<>0 then		 
				valueofloss:= valueofloss + techniques[t,dest,produces[t]]*x[t]*epsilon*L[produces[t]];
		     
		     marginalgain := valueofincrease - valueofloss
	     end
     end;
     function mpp( s :technique; d :resource):volume;
    (*! This computes the marginal  physical product or the output of s with respect to input d   *)
     
     begin { mpp}
      
        (*! check if technique s uses resource d *)
        if techniques[s,src,d] =0 then
        begin
         
           mpp:= - maxint
        end
        else begin   
	         mpp := techniques[s,dest,produces[s]]/techniques[s,src,d];
	       
	     end
     end;
     function mostProductiveTechniqueFor(e:resource):technique;
     (*! This determines within   which technique resource e produces the most of. *)
     var v:volume;j,s :technique;
     begin
	    v:=0.0;
        s:=resource0;
        for j:= alpha to techniqueN do
        if techniques[j,src,e]<0 then { it uses this input }
        begin
            if v< techniques[j,dest,produces[j]]*x[j]/(  -techniques[j,src,e] )then
			begin
				v:=techniques[j,dest,produces[j]]*x[j]/(  - techniques[j,src,e]);
				 s:=j;
				 
			end;
			 
        end;
        

	    mostProductiveTechniqueFor:=s;
     end;
    procedure ComputeTotalsEtc;
	(*! Work out how much is being produced, which resource is  being produced least and which resources outputs are equals to this, *) 
     var d :array[resource]of real;i,j:integer;
     begin
		    totals:=0;
		   for i := alpha to techniqueN do begin
		      totals[produces[i]] := totals[produces[i]]+ (techniques[i ,dest,produces[i]]) * x[i];
		      
		   end;
            
			(*! totals now has a vector of how much of each resource is produced at the end relative to the target *)
			weightedTotalOutput := L . totals;
			leastProducedOutput := \min (totals -target);
			(*! Find the resource that is least produced or most in shortage *)
			for s:= resource0 to resourceM do 
				if (totals[s]-target[s])=leastProducedOutput then least:=s;
			(*! Find the ones on the plan ray *)
			d:=(totals - target -leastProducedOutput) ;
			equated := ord( (d<epsilon)and(d> -epsilon) ); 
			{writeln('t',totals,'d',d);
			writeln('e',equated,'x',x);
			writeln('least',least);
			write('press return');readln;}
     end;
   
     function findScoop(var s:technique):resource;
     (*!  search all output who are not yet reaching target to find a pair of  input and output where
     shifting input time to producing that output would produce the best result.  Return the input as the function
     result and the output in the parameter s *)
     var
	 gain:volume;m:resource;j:technique;
	 d,Scoop:resource;  
     begin {find Scoop} 
		Scoop := A;
		gain := -maxint ;
		s:=alpha;
		for d:= resource0 to resourceM do 
		begin
		    if startingresource[d]>0 then{ d iterates over inputs }
			for j:=  alpha to techniqueN do			    
				if equated[produces[j]]>0 then
				begin				   
					m:=produces[mostProductiveTechniqueFor(d)]; 
					
					if marginalgain(j,d)>gain then
					begin
						if equated[m]<1 then
						begin 
							gain:= marginalgain(j,d);
							
							Scoop:= d; 
							s:= j;
						end;
					end;
				end;
		end;
		
		findScoop:= Scoop;
     end;
     function somePositiveTargetsStillZero:boolean;
     (*! returns true if some products which should have positive output are still zero *)
     begin
		somePositiveTargetsStillZero:= \ OR ((totals<=0)AND(target>0))
	 end;
begin {Kantorovich algorithm}
	L:=1; { initialise multipliers}
	ok:=false; count:=0;f:=0.3;
        (*! Iterate the following steps until we have a satisfactory answer. *)
	while not ok do 
	begin
	    available:=startingresource;
		x:=0  ; {default assumption }
		(*!     Use the L    to get a standardised performance for each process. *)
		standardisedoutput :=  (techniques[iota 0,dest]  *   L);
		(*! For each input find the technique for which it has the best performance *)
		for e := resource0 to resourceM do
		if available[e]>0 then { test it is an input }
		begin
		(*! find the best performance of the input on any technique *)
		    outputs:=0;
		    for s:= alpha to techniqueN do { try every technique}
		      if techniques[s,src,e] <0 then { it uses resource e }
		        outputs[s] := \+ standardisedoutput[s];		 
			best := \max outputs;		 
			(*! set each machine to work on the technique it is best at *)
			for  s := alpha to techniqueN do 
				if outputs[ s]=best then 
				begin
				(*! set this technique to have an intensity sufficient
				    to use all of the resource e that is available 
				    subject to the availability of the other resources technique s needs *)
				    achievableintensities := - available /(techniques[s ,src]-epsilon); {$\epsilon$ term prevents divide by 0 }
				    targetintensity := \min achievableintensities[e];
				   
				    (*! We find the lowest of the intensities that can be achieved with remaining resources. *)
				    x[s] := x[s]+ targetintensity;{ add it into x }
				end;
			  
		end;
		totals:=0;
		for k := alpha to techniqueN do
		totals[produces[k]] := totals[produces[k]]+ (techniques[k ,dest,produces[k]]) * x[k];
		greatestProducedOutput := \max totals;
		leastProducedOutput := \min totals;
		
		 
		if somePositivetargetsStillZero  then
		begin
			(*! check if any resource has a zero output and raise its value if it has*)
			for s:= resource0 to resourceM do 
			    if target[s] >0 then { s iterates over outputs }
				if totals[s]<greatestProducedOutput then 
					L[s]:= L[s]*(1.02+(ord(s)/(10)))
		end
		else ok:=true;
	end;
	count:=0;
	(*! At this point our estimate of the resolving multipliers is accurate enough to ensure
            that some of each output is now being produced, but we have not yet met the requirement
            that the targer amount of each output must be produced. We now try to get a more precise
            estimate of the resolving multipliers and in the process we adjust the amounts
            of each output being made. It is important to note at this point that any further
            adjustments must come by de-specialising some of the inputs so that they are used
            more than one output type. The resolving multipliers have until now been  used to weight the outputs
            of different  types in order to assign each input to the output it is best suited to.
            If an input is no longer specialised, that is if it produces more than one routput, then
            the weights must be such that it is no longer {\em best} at one particular output type.
            The multipliers must be set so that the marginal weighted output of the resource on
            any of the outputs on which it is employed are the same. Thus if a machine $k$ is employed
            on two resources $i,j$ then {\sf standardisedoutput[k,j]=standarisedoutput[k,i]}. 

		In turn this implies that for any machine that is employed to make two resources
            the ratio of the resolving multipliers must be the inverse of the ratio of the techniques.

	    The algorithm will work output a time bringing ever more  outputs into equality with their targets.
            We define the set of resources whose outputs has been brought into equality with target as the equated set.

	    For those outputs in the equated set, the resolving multipliers of the outputs will have been corrected
            so that for any machine moving more than one resource they stand in inverse ratio to
            that machine's productivity.
	 *)
	 
	computeTotalsEtc; 
	repeat 			 
			(*! Find which input non in the equated set is 2nd best at producing this output under current resolving multipliers. Call this input Scoop. *)
		  Scoop := findScoop(least);
		  
			 m := mostProductiveTechniqueFor(Scoop);
			  
		 	(*! Adjust the resolving multiplier ratio between Scoops resource and the least produced resource to
                            ratio of Scoops techniques. *)
          
			L [produces[m]]:=L [produces[least]]*mpp(least,scoop)/mpp(m,scoop);
			 (*! It is now necessary to reduce the ouput of scoop on scoops output and increase it
                            on the least produced output. It is necessary to compute how much to reduce scoops resource by.
                            The resolving multipliers give us substitution ratios between different resource outputs.
                            Suppose that we want to reduce output of resource $m$ by one unit and increase the output
                            of resource $j$, the increase in $j $ we get is
				 $$ \Delta_j=\frac{L_m}{L_j}$$.
                            If we want to reduce the output of $i$  and we have two other resources $j,k$ which
                            we want to increase  equally then we have 
				$$\Delta_k=\Delta_j$$ 
				where $\Delta_x$ means change in $x$ and
                            $$ - \Delta_m L_m = \Delta_k L_k + \Delta_j L_j=\Delta_k(L_j+L_k)$$.
				Thus $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m} $$
                            Let $C_m, C_j, C_k$ be the current outputs of each resource;  given that $C_j=C_k$ 
                            we have to chose the $\Delta$s so that $$ C_m+\Delta_m = C_j+\Delta_j=C_k+\Delta_k$$
                            It follows that $$C_m-\Delta_k\frac{L_j+L_k}{L_m} =C_k+\Delta_k$$
				and$$C_m-C_k=\Delta_k\frac{L_j+L_k}{L_m} +\Delta_k=\Delta_k(1+\frac{L_j+L_k}{L_m})$$
			   	so$$\Delta_j= \frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}$$
			We next compute the reduction to be made in resource $m$ from the formula $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m}$$
			substitiuting we get
			$$\Delta_m= -\frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}(\frac{L_j+L_k}{L_m})$$
			    Translating this to the variables used in the program we have:
                         *)
			j:=produces[least]; 
						
			lambda :=  (L.equated)/L[produces[m]];
			(*! we have used the cross product with the equated set, represented as a vector of 0 or 1 to sum over the equated set.*)
			
            deltam :=  (-1)*lambda*(totals[produces[m]]-totals[j])/(1+lambda) ; 
			
                        (*!  Note that in the line above we are generalising the term $L_j+L_k$ to an aribtrary number
                             of multipliers ( 1 or 2  in this program ) by computing the inner product between the
                             equated vector and the multipliers. This works because the equated vector has a 1 for
                             all resources in the equated set.
			    We now compute the change in duration that Scoop spends on its best resource (dx) by 
                            scaling deltam by Scoops technique output for resource m.
                         *)
           
		    dx := deltam/ (- mpp(m,scoop)); 	   
		  
		  x[m ]:=x[m ] +dx;
                        (*! reallocate this time to Scoops best resource in the equated set which we will now call $j$ *)
          
            best:= -maxint;
            jt:=alpha;
          
          
			for st:=alpha to techniqueN do 
				if mpp(st,scoop) * equated[produces[ st]  ]>best then
				begin   
					best:= mpp(st,scoop);
					jt:=st;
			
             
				end;
		 
			x[jt ]:= x[jt ]-dx;
			
            
			computeTotalsEtc;	  
			count:=count+1;
			
	until ((\+ equated) = 3) or (count>20);
	writeln('answer arrived at after ',count,' trys');
	writeln ('technique intensities '   );
	writeln ( x   );
	writeln('flows');writeln(totals);
end {Kantorovich algorithm}.
