 (*! What follows is a library written in Vector Pascal.
 The text in roman font that follows
are comments. The program code is generally in san-serif font, and
the whole, is in the literate programming output format generated
by the compiler. *)
Unit Kantor;

(*!
*
The library provides a procedure which generalises the algorithm of
Kantorovich that he first presented in the context of excavators
working on soils of different types, so that it will work for
a general plan optimisation for the whole economy
by his method of resolving multipliers. It starts out from
the data provided in an i/o Table.
In the table the {\em techniques} for for different production techniques are shown
in italics. In this context a technique  means the
expected output per year of the technique. In the case of Kantorovich's
original algorithm the techniques were represented as a matrix of
outputs of different soils and machines. In an input output table
a technique is conventionally represented by a column (Leontief notation)
or a row (Sraffa notation) of a matrix.

We can convert the Kantorovich table $$
\begin{array}{ccc}105& 107& 64\\
						 56& 66&38\\
						 56& 83& 53\end{array}$$
Where each row corresponds to a soil type and each column to
an excavator into an extended Sraffa format as
$$
\begin{array}{cccccc}
105&0&0&-1&0&0\\
107&0&0& 0&-1&0\\
64&0&0& 0&0&-1\\
0&56&0& -1&0&0\\
0&66&0& 0&-1&0\\
0&38&0&  0&0&-1\\
0&0&56& -1&0&0\\
0&0&83& 0&-1&0\\
0&0&53&  0&0&-1
\end{array}
$$

where the first 3 columns correspond to the outputs of different soils
and the next 3 to the outputs of excavators. The outputs of excavators
are negative since a given process uses up an excavator.

In practice we split this into two distinct tables one the {\tt dest}
table which lists the outputs and the other the {\tt src} table which
lists the inputs.
Thus:

{\tt dest}
$$
\begin{array}{cccccc}
105&0&0&0&0&0\\
107&0&0& 0&0&0\\
64&0&0& 0&0&0\\
0&56&0& 0&0&0\\
0&66&0& 0&0&0\\
0&38&0&  0&0&0\\
0&0&56& 0&0&0\\
0&0&83& 0&0&0\\
0&0&53&  0&0&0
\end{array}
$$
{\tt src}
$$
\begin{array}{cccccc}
0&0&0&-1&0&0\\
0&0&0& 0&-1&0\\
0&0&0& 0&0&-1\\
0&0&0& -1&0&0\\
0&0&0& 0&-1&0\\
0&0&0&  0&0&-1\\
0&0&0& -1&0&0\\
0&0&0& 0&-1&0\\
0&0&0&  0&0&-1
\end{array}
$$
Why reorganise like this?

Because the general economic planning problem involves the production
and consumption of many of the same goods. Energy is both used and
produced, for example. It is thus convenient to show the inputs and
outputs in two different tables.

The pair composed of a destination row and the corresponding src
row we term a technique.

In principle we could have a problem with joint production, but for
now we are sticking to the simpler single product per technique
problem. Thus each row of the destination column contains only one
non zero element.

The objective was given, by Kantorovich  in a final  column  of his table,
in our format we would have
row vector of resources available at the start of the period.

Kantorovich's algorithm demanded equal quantities of all outputs
which is not the general case. We keep the Kantorovich approach
within this library, which means that the units of measurement
of all inputs must be scaled in order to result in an output that
is the same, in these units of measurement, for all products.

Thus if we wanted to produce 2 cubic meters of soil type 1 relative
to every cubic meter of soil type 2 and 3 we use 2 cubic meter as the
standard unit for 1.

We handle this by dividing the appropriate column of the technology
matrix by 2 and similarly scaling any corresponding entries for
the available resources.
More generally, given a targets vector we divide all other row vectors
in the problem by this.
$$
\begin{array}{cccccc}
 20000&20000&20000&0&0&0
\end{array}
$$
And an initial resource vector of
$$
\begin{array}{cccccc}
0&0&0&1&1&1
\end{array}
$$
indicating we have 1 machine of each type.

My first step in developing the programme in this new format was to
take the programme that I had previously published in Intervention, and
to reformulate it in this more general form. The aim is to check that
given the new data format the algorithm still gives the same answers
as Kantorovich. Following on from that we can provide a more
general version that will work on arbitrary sized input output
tables.

Following the columns of the techniques Kantorovich gives the optimal allocation
of machine times to activities to minimise overall time taken to do the digging.
The program will reproduce this result by applying his method of resolving
multipliers or objectively determined valuations.
We first introduce our domain of discourse resources, techniques and units of  usage and output. We simply
number the resources 0 to 5 with resources in this case including soil
and machines.
*)
interface
const
	  printenabled=false;
      technique1=1;{ techniques go from technique1 to techniqueN}
	  epsilon =0.001;
	  resource0=0;  { first entry in US IO table is resource0 }
	                  { resources or products from resource0 to resourceM }
	  dest=0;
	  src=1;
	  growthconst=0.1;
	  annealrate=0.99;
type
	 optimizer=(kantorovichs,pauls);
     flow = dest..src;
     resource = integer;
     units =real;
     intensity = real;
     ODV = real; { Kantorovich's Objectively Determined Valuations}

     technique = integer;
     resourcevec(resourceM:resource) = array[resource0..resourceM ] of units;{ last entry of US IO table is resourceM }
     (*! A technology is defined as a flow of  src inputs to a flow of dest outputs,
     with each flow being a vector. *)

     technologytable(techniqueN,resourceM:integer) = array [technique1 .. techniqueN,dest..src,resource0 .. resourceM] of units;
     ODVvec=resourcevec;

     intensityvec=vector;
     volume=real;
var growthfactor:real;
    optimiser:optimizer;
procedure  kantorovichMethod(var techniques:technologytable; var target,startingresource:resourcevec; var L:ODVvec; var X:intensityvec);

            (*! The method must be supplied with a technology table and a vector of targets and available resources.
            It outputs  a vector of resolving multipliers L and an vector of technique intensities X
            which specify how much each technique will be used. *)

IMPLEMENTATION


type productionIndex (techniqueN:integer)=array[technique1 .. techniqueN] of resource;
     pproductionIndex = ^ productionIndex;

procedure IndexedMethod( var produces:productionindex;var techniques:technologytable; var targets,startingresource,L:ODVvec; var X:intensityvec);
(*! The indexed method is a nested procedure which will work using
an index of what each technique produces. Many techniques may produce same output. *)

(*! The algorithm manipulates  a vector x which will encode  the time  for which each technique
    is operated, this corresponds in Kantorovich's case to the time a given type of machine spends on a given soil.
    Later it may have a more general impliation. Next  L, a vector of objectively determined valuations  of resources\footnote{
    Presumably Kantorovich labels his multipliers L because of their similarity to Lagrangian multipliers}.

    The standardised output of each technique  is obtained by applying resolving multipliers
    to the outputs  types it produces.
*)
type outputmatrix(techniqueN,resourceM:integer) =array [technique1 .. techniqueN,resource0 .. resourceM] of units;
var
    targetintensity:real;
    achievableintensities:^resourcevec;
    dx: intensity;
    standardisedoutput:^ outputmatrix;
    available: ^resourcevec;
    outputs : ^ intensityvec;
    totals: ^odvvec;
    ok:boolean;
    greatestProducedOutput,leastProducedOutput ,deltam: real;
    best:real;
    e, e2,Scoop,
    A:resource; {first input resource }
    r,s, j: resource;
    weightedTotalOutput:real;{Output weighted by L}
    lambda,f,t:real;
    least,m,jt,st:technique;
    count,k:integer;
    equated: ^odvvec;
    Function findfirstinput:resource;
    var t:technique; r ,f: resource;
    begin
        f:=techniques.resourceM;
		for t:= technique1 to techniques . techniqueN do
		  for r:= f downto 0 do
		      if techniques[t,src,r] >0 then f:=r;
		findfirstinput:=f
	end;{Find first input} 

    procedure claimHeapSpace;
    var maxtechnique,maxresource:integer;
    begin{claimHeapSpace;} 
        maxtechnique:=techniques . techniqueN;
        maxresource:=techniques.resourceM;
		new(achievableintensities,maxresource);
		new(standardisedoutput,maxtechnique,maxresource);
		new(available,startingresource.resourceM);
		new(outputs,maxtechnique);
		new(totals,maxresource);
		new(equated,maxresource);
		A:= findfirstinput;
	end;{claimHeapSpace}
	procedure freeheapspace;
	begin
		dispose(achievableintensities);
		dispose(standardisedoutput);
		dispose(available);
		dispose(outputs);
		dispose(totals);
		dispose(equated);
	end;

    function marginalgain( s :technique; d :resource):volume;
    (*! This computes the marginal gain,   of a
        small shift of the resource  d  time to the specified technique type s. We compute the effect of
        multiplying all current time allocations to $1-\epsilon$ whilst increasing the allocation of time
        to resource s by $\epsilon$.   *)

     var currentoutput, currentuse, shift, marginalIncreaseinS:real;
         valueofincrease,valueofloss:real;
         t:technique;
     begin { marginal gain}

        (*! check if technique s uses resource d *)
        if techniques[s,src,d] =0 then
        begin
           marginalgain:= - maxint;
        end
        else
        begin
	     (*! get total current use of resource d *)
	        currentuse := \+ x* techniques[iota 0, src, d];{ current use is negative !}
	        shift:= epsilon * currentuse;                  { shift will be negative too}
	        marginalIncreaseInS:= shift* techniques[s,dest,produces[s]]/techniques[s,src,d];{ marginal increase will be +ve, since we divide by a negative}


			valueofincrease:= marginalIncreaseInS * L[produces[s]];
			(*! we next have to work out the total value of output currently produced in all sectors by d *)
			valueofloss:=0;
			for t:= technique1 to techniques . techniqueN do
		 	 if techniques[t,src,d]<>0 then	{ if it uses d}
				valueofloss:= valueofloss + techniques[t,dest,produces[t]]*x[t]*epsilon*L[produces[t]];

		     marginalgain := valueofincrease - valueofloss
	     end
     end;{marginalgain}

     function mpp( s :technique; d :resource):volume;
    (*! This computes the marginal  physical product or the output of s with respect to input d   *)

     begin { mpp}
        (*! check if technique s uses resource d *)
        if techniques[s,src,d] =0 then
        begin
             mpp:= - maxint;
        end
        else
         begin
	         mpp := techniques[s,dest,produces[s]]/techniques[s,src,d];
	     end
     end;{mpp}

     function mostProductiveTechniqueFor(e:resource):technique;
     (*! This determines within   which technique resource e produces the most of. under current valuations. *)
     var v:volume;j,s :technique;
     begin
	    v:=0.0;
        s:=technique1;
        for j:= technique1 to techniques . techniqueN do
        if techniques[j,src,e]<0 then { it uses this input }
        begin
            if v< L[produces[j]]*techniques[j,dest,produces[j]]*x[j]/(  -techniques[j,src,e] )then
			begin
				v:=L[produces[j]]*techniques[j,dest,produces[j]]*x[j]/(  - techniques[j,src,e]);
				 s:=j;
			end;
        end;

        mostProductiveTechniqueFor:=s;
     end;{mostProductiveTechniqueFor}

    procedure ComputeTotalsEtc;
	(*! Work out how much is being produced, which resource is  being produced least and which resources outputs are equals to this, *)
     var i,j:integer;least:real;
     begin
		    totals^:=0;
		   for i := technique1 to techniques . techniqueN do
		   begin
		      totals^[produces[i]] := totals^[produces[i]]+ (techniques[i ,dest,produces[i]]) * x[i];
		   end;

			(*! totals now has a vector of how much of each resource is produced at the end relative to the target *)
			weightedTotalOutput := L . totals^;
			leastProducedOutput := \min (totals^ );
			(*! Find the resource that is least produced or most in shortage *)
			for s:= resource0 to techniques.resourceM do
				if (totals^[s])=leastProducedOutput then least:=s;
			(*! Find the ones on the plan ray *)
			equated^:=0;least:=maxint;
			for i:= resource0 to techniques.resourceM do
				if targets[i]>0 then
			   if totals^[i] < least then least:= totals^[i];
			   
			for i:= resource0 to techniques.resourceM do
				if targets[1]>0 then
				if (totals^[i]-least)<epsilon then equated^[i]:=1;
			
		
			equated^ := ord( ((totals^  -least) <epsilon)and((totals^  -least)  > -epsilon) ); {}
	{writeln('t',totals^);
			writeln('e',equated^);writeln('x',x);
			writeln('least',least);
			{write('press return');readln;{}


     end;

     function somePositiveTargetsStillZero:boolean;
     (*! returns true if some products which should have positive output are still zero *)
     begin
		somePositiveTargetsStillZero:= \ OR ((totals^  <=0)AND(targets>0));
	 end;
	 function numpostargets:integer;
	 begin
		numpostargets := \+ (1 AND (targets>0))
	 end;
	 procedure renormaliseMultipliers;
	 (*! In order to prevent uncontrolled growth or shrinkage in multipliers, express them with respect to their own average.*)
	  var av:real;
	 begin
	    av:=\+ L / (1+techniques.resourceM);{ find the average}			
		L:= L/av;{ renormalise L}
	 end;
	 procedure LowerNonEquatedMultipliers;
	 begin
		L:=L*(equated^ *growthfactor +1 -growthfactor);		{those in the equated set are unchanged by this others are reduced}
		renormaliseMultipliers;
	 end;
	 procedure LowerNonEquatedIntensities;
	 begin
		X :=X*(equated^[produces] *growthfactor +1 -growthfactor);
	 end;
	 procedure getavailability;
	 var i:integer;
	 begin
	 
	    available^:=startingresource;
	    for i:= technique1 to techniques.techniqueN do
	     available^:= available^+ techniques[i,src]*X[i];
	 end;
	 procedure RedistributeResources;
	 (*! move the free resources to the place where they are most valuable *)
	 begin
	 (*!     Use the L    to get a standardised performance for each process. *)

	     standardisedoutput ^:=  (techniques[iota 0,dest]  *   L);
	     getavailability;
		(*! For each input find the technique for which it has the best performance *)
		for e := resource0 to techniques.resourceM do
		if available^[e]>0 then { test it is an input }
		begin
			(*! find the best performance of the input on any technique *)
		    outputs^:=0;
		    for s:= technique1 to techniques. techniqueN do { try every technique}
		      if techniques[s,src,e] <0 then { which uses resource e }
		      begin
		        t:=0;
		        for r:= resource0 to techniques.resourceM do
		        t := t+ standardisedoutput^[s,r];
		        outputs^[s]:=t;
		      end;
			best := \max outputs^;
			(*! allocate each resource to  the technique it is best at *)
			for  s := technique1 to techniques. techniqueN do
				if outputs^[ s]=best then
				begin
				(*! set this technique to have an intensity sufficient
				    to use 50 percent of the resource e that is available
				    subject to the availability of the other resources technique s needs. The achievable intensities
				    specifies how much more output this technique could produce subject to the constraint
				    of each type of resource's availability. Thus it is indexed by the available resource types. *)
				    achievableintensities^ :=if techniques[s,src]=0.0 then maxreal
				                             else  - available^ /(techniques[s ,src]);

				    targetintensity :=0.5* \min achievableintensities^;

				    (*! We find the lowest of the intensities that can be achieved with remaining resources. *)
				    x[s] := x[s]+ targetintensity;{ add  it into x }
				    (*! update the available resources , we add the source weighted by intensity, note that these numbers
				    will be negative and will reduce availability *)
				    available^:= available^ +(targetintensity * techniques[s,src]);
				end;

		end;
	
	 end;
	  procedure rescaleall;
	 (* this, having found the best ratio of intensities, scales them all to the 
	 limit of the available resources by finding which is the rate limiting resource *)
	 var usage:^resourcevec;lowestsurplus:real;
	 begin
	  getavailability;
	  new(usage,techniques.resourceM);
	  usage^:= startingresource -available^;
	  lowestsurplus:= \min (startingresource/usage^);
	  X:=X*lowestsurplus;dispose(usage);
	  end;
	
	 procedure PaulsOptimiser;
	 (* ! this is an alternative algorithmic step for finding the optimal resolving multipliers
	 once an initial set has been found.
	 
	 Its basic steps are:
	 \begin{enumerate}
	 \item Lower the resolving multipliers of the non equated outputs relative to the equated ones.
	 \item Lower the intensities (X) of all processes producing non equated outputs.
	 \item Redistribute the free resources to the processes whose outputs are in the equated set.
	 \end{enumerate}
	 *)
	 begin
	     
	     LowerNonEquatedMultipliers;
	     LowerNonEquatedIntensities;
	     RedistributeResources;
	     rescaleall;
	     growthfactor:=growthfactor*annealrate;
	     
	 end;
	  
	  
	 procedure KantorovichsOptimiser;
	 (*! This optimising step is called repeatedly once an initial set of resolving multipliers has been found. It is the 
	 one originally proposed by Kantorovich himself. *)
	 
	     function findScoop(var s:technique):resource;
	     (*!  search all outputs which are not yet reaching target to find a pair of  input and output = produces[s] where
	     shifting input time to producing that output would produce the best result.  Return the input as the function
	     result and the technique to use in the parameter s *)
	     var
		 gain:volume;m:resource;j,t:technique;
		 d,Scoop:resource;
	     begin {find Scoop}
			Scoop := A;m:=0;
			gain := -maxint ;
			s:=technique1;
			for d:= resource0 to techniques.resourceM do
			begin
			    if startingresource[d]>0 then
			    begin
			    (*! that is to say we iterate over all goods used as inputs *)
			        m:=produces[mostProductiveTechniqueFor(d)]; 	{ m stands for Main output of d }
	
					for j:=  technique1 to techniques . techniqueN do
						if equated^[produces[j]]>0 then
						(*!find a technique j, whose output is doing ok that potentially could have resource reallocated to make d *)
						begin
							if marginalgain(j,d)>gain then
							begin
							    (*! if this is the best marginal gain so far, record the technique j and the input d *)
								if equated^[m]<1 then
								begin
									gain:= marginalgain(j,d);
								{	writeln('gain',gain,',j',j,',d',d,',m',m);}
									Scoop:= d;
									s:= j;
								end;
							end;
						end;
				end;
			end;
			{writeln('best gain',gain,',s=',s,',scoop=',scoop,',best output from scoop',m);}
			findScoop:= Scoop;
	     end;
	 begin
	 
		(*! Find which input  is 2nd best at producing output least in the non equated set  under current resolving multipliers. Call this input Scoop.
		The parameter least is an output of the function call. *)
		    Scoop := findScoop(least);
			 m := mostProductiveTechniqueFor(Scoop);
		 	(*! Adjust the resolving multiplier ratio between Scoops resource and the least produced resource to
                            ratio of Scoops techniques. *)
			L [produces[m]]:=L [produces[least]]*mpp(least,scoop)/mpp(m,scoop);
			 (*! It is now necessary to reduce the ouput of scoop on scoops output and increase it
                            on the least produced output. It is necessary to compute how much to reduce scoops resource by.
                            The resolving multipliers give us substitution ratios between different resource outputs.
                            Suppose that we want to reduce output of resource $m$ by one unit and increase the output
                            of resource $j$, the increase in $j $ we get is
				 $$ \Delta_j=\frac{L_m}{L_j}$$.
                            If we want to reduce the output of $i$  and we have two other resources $j,k$ which
                            we want to increase  equally then we have
				$$\Delta_k=\Delta_j$$
				where $\Delta_x$ means change in $x$ and
                            $$ - \Delta_m L_m = \Delta_k L_k + \Delta_j L_j=\Delta_k(L_j+L_k)$$.
				Thus $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m} $$
                            Let $C_m, C_j, C_k$ be the current outputs of each resource;  given that $C_j=C_k$
                            we have to chose the $\Delta$s so that $$ C_m+\Delta_m = C_j+\Delta_j=C_k+\Delta_k$$
                            It follows that $$C_m-\Delta_k\frac{L_j+L_k}{L_m} =C_k+\Delta_k$$
				and$$C_m-C_k=\Delta_k\frac{L_j+L_k}{L_m} +\Delta_k=\Delta_k(1+\frac{L_j+L_k}{L_m})$$
			   	so$$\Delta_j= \frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}$$
			We next compute the reduction to be made in resource $m$ from the formula $$\Delta_m= -\Delta_k\frac{L_j+L_k}{L_m}$$
			substitiuting we get
			$$\Delta_m= -\frac{C_m-C_j}{1+\frac{L_j+L_k}{L_m}}(\frac{L_j+L_k}{L_m})$$
			    Translating this to the variables used in the program we have:
                         *)
			j:=produces[least];

			lambda :=  (L.(equated^))/L[produces[m]];
			(*! we have used the cross product with the equated set, represented as a vector of 0 or 1 to sum over the equated set.*)
            deltam :=  (-1)*lambda*(totals^[produces[m]]-totals^[j])/(1+lambda) ;
                        (*!  Note that in the line above we are generalising the term $L_j+L_k$ to an aribtrary number
                             of multipliers ( 1 or 2  in this program ) by computing the inner product between the
                             equated vector and the multipliers. This works because the equated vector has a 1 for
                             all resources in the equated set.
			    We now compute the change in intensity that Scoop spends on its best resource (dx) by
                            scaling deltam by Scoops technique output for resource m.
                          *)
		  dx := deltam/ (- mpp(m,scoop));
                        (*! reallocate this time to Scoops best resource in the equated set which we will now call $j$ *)
            best:= -maxint;
            jt:=technique1;


			for st:=technique1 to techniques  . techniqueN do
				if mpp(st,scoop) * equated^[produces[ st]  ]>best then
				begin
					best:= mpp(st,scoop);
					jt:=st;
				end;
			
		    x[m ]:=x[m ] +dx;
			x[jt ]:= x[jt ]-dx;
	 end;
	 var av:real;
begin {Kantorovich algorithm}
    claimHeapSpace;

	L:=1; { initialise multipliers}
	ok:=false;
	count:=0;
	f:=0.3;
        (*! Iterate the following steps until we have a satisfactory answer. *)
	while not ok do
	begin
	    available^:=startingresource;
		x:=0  ; {default assumption }
		redistributeResources;
	 	totals^:=0;
		(*! Compute the total output of each product in value terms *)
		for k := technique1 to techniques . techniqueN do
		totals^[produces[k]] := totals^[produces[k]]+ (techniques[k ,dest,produces[k]]) * x[k]*L[produces[k]];
		greatestProducedOutput := \max totals^;
		leastProducedOutput := \min totals^;
		if somePositivetargetsStillZero  then
		begin
			(*! check if any resource has a zero output and raise its value if it has*)
			for s:= resource0 to techniques . resourceM do
			    if targets[s] >0 then { s iterates over outputs }
				    if totals^[s]<=0 then L[s]:=L[s]*(1.0+growthfactor*(random and 7)/8 + growthfactor);
			 renormaliseMultipliers;
		end
		else ok:=true;
		count:=count+1;
		{writeln(X,L/L[0],count);}
		growthfactor:=growthfactor*annealrate;
		if count >techniques.techniqueN*10000  then
		begin
		 writeln('after ',count,' trys could not find initial resolving multipliers');
		 halt(999);
		end;
	end;
	{writeln('initial multipliers found after ',count,' trys');
	writeln(L);}
	count:=0;
	(*! At this point our estimate of the resolving multipliers is accurate enough to ensure
            that some of each output is now being produced, but we have not yet met the requirement
            that the targer amount of each output must be produced. We now try to get a more precise
            estimate of the resolving multipliers and in the process we adjust the amounts
            of each output being made. It is important to note at this point that any further
            adjustments must come by de-specialising some of the inputs so that they are used
            more than one output type. The resolving multipliers have until now been  used to weight the outputs
            of different  types in order to assign each input to the output it is best suited to.
            If an input is no longer specialised, that is if it produces more than one routput, then
            the weights must be such that it is no longer {\em best} at one particular output type.
            The multipliers must be set so that the marginal weighted output of the resource on
            any of the outputs on which it is employed are the same. Thus if a machine $k$ is employed
            on two resources $i,j$ then {\sf standardisedoutput[k,j]=standarisedoutput[k,i]}.

		In turn this implies that for any machine that is employed to make two resources
            the ratio of the resolving multipliers must be the inverse of the ratio of the techniques.

	    The algorithm will work output a time bringing ever more  outputs into equality with their targets.
            We define the set of resources whose outputs has been brought into equality with target as the equated set.

	    For those outputs in the equated set, the resolving multipliers of the outputs will have been corrected
            so that for any machine moving more than one resource they stand in inverse ratio to
            that machine's productivity.
	 *)
    growthfactor:=growthconst;
	computeTotalsEtc;
	repeat
			case optimiser of
			kantorovichs:	KantorovichsOptimiser;
			pauls:			PaulsOptimiser;
			end;
			computeTotalsEtc;
			count:=count+1;

	until ((\+ equated^) = numpostargets) or (count>(500))or (growthfactor<epsilon);
{	writeln('second pass terminated after ', count);}
	rescaleall;
	freeHeapSpace;
end {Indexed method};



procedure  kantorovichMethod(var techniques:technologytable; var target,startingresource,L:ODVvec; var X:intensityvec);
var index:^productionindex;t:technique;r:resource;targets:^odvvec;
begin { Kantorovich method }
   if printenabled then
   begin
	 writeln(techniques:12:2,target, startingresource);
   end;
   new (index, techniques . techniqueN);
   
   growthfactor:=growthconst;
   new (targets,target.resourceM);
   targets^:= if target=0.0 then 0.0 else target/target[0];
   for t:= technique1 to techniques . techniqueN do
    for r:= resource0 to techniques.resourceM do
     if techniques[t,dest,r]>0 then
     begin
      index^[t]:=r; 
      end;
   (*! prescale by targets *)
    for t:= technique1 to techniques . techniqueN do
    begin
    
     if target[index^[t]]<>0 then 
		techniques[t]:= techniques[t] / targets^ [index^[t]] ;
   end;
    for r:=resource0 to target.resourceM do 
	  if (target[r]<>0) then
	    startingresource[r]:=startingresource[r] /targets^[r ] ;
   indexedMethod(index^,techniques, targets^, startingresource,L, X);
   (*! return scale of the input data to its original value *)
    for t:= technique1 to techniques . techniqueN do
    begin
     if target[index^[t]]<>0 then 
		techniques[t]:= techniques[t] * targets^ [index^[t]] ;
   end;
    for r:=resource0 to target.resourceM do 
	  if (target[r]<>0) then
	    startingresource[r]:=startingresource[r] *targets^[r ] ;
   dispose(index);dispose(targets);
end;{Kantorovich method }


Begin { unit initialisation}
growthfactor:=growthconst;
optimiser:=kantorovichs;
end {unit kantor }.
